# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import scipy as sp
import scipy.linalg
import scipy.stats
import limix.deprecated.utils.preprocess as preprocess
from limix.utils.preprocess import covar_rescaling_factor 
from limix.utils.util_functions import vec 
from limix.core.covar import *
from limix.core.gp import *
from limix.core.mean import *
import pdb
import time
import copy
import warnings

class VarianceDecomposition:
    """
    Variance decomposition module in LIMIX
    This class mainly takes care of initialization and eases the interpretation of complex variance decompositions

    vc = varianceDecomposition(Y) # set phenotype matrix Y [N,P]
    vc.addFixedEffect() # set intercept
    vc.addRandomEffectTerm(K=K) # set genetic random effect with genetic kinship as sample covariance matrix [N,N]
    vc.addRandomEffectTerm(is_noise=True) # set noisy random effect
    vc.optimize() # train the gaussian process
    vc.getTraitCovar(0) # get estimated trait covariance for random effect 1 [P,P]
    vc.getVarianceComps() # get variance components of the different terms for different traits as [P,n_randEffs]
    """

    def __init__(self, Y, standardize=False):
        """
        Args:
            Y:              phenotype matrix [N, P]
            standardize:    if True, impute missing phenotype values by mean value,
                            zero-mean and unit-variance phenotype (Boolean, default False)
        """
        #check whether Y is a vector, if yes reshape
        if (len(Y.shape)==1):
            Y = Y[:,sp.newaxis]

        #create column of 1 for fixed if nothing provided
        self.N       = Y.shape[0]
        self.P       = Y.shape[1]
        self.Nt      = self.N*self.P
        self.Iok     = ~(sp.isnan(Y).any(axis=1))

        #outsourced to handle missing values:
        if standardize:
            Y = preprocess.standardize(Y)

        # set properties
        self.Y = Y
        self.n_randEffs  = 0
        self.noisPos = None

        # fixed and ranodm effects 
        self.sample_covars = []
        self.sample_cross_covars = []
        self.trait_covars = []
        self.sample_designs = []
        self.sample_test_designs = []
        self.trait_designs = []

        # desinc
        self._desync()

        #TODO: check if necessary
        self.Ntest = None

    def setY(self,Y,standardize=False):
        """
        Set phenotype matrix

        Args:
            Y:              phenotype matrix [N, P]
            standardize:    if True, phenotype is standardized (zero mean, unit variance)
        """
        assert Y.shape[0]==self.N, 'VarianceDecomposition:: Incompatible shape'
        assert Y.shape[1]==self.P, 'VarianceDecomposition:: Incompatible shape'


        if standardize:
            Y = preprocess.standardize(Y)

        #check that missing values match the current structure
        assert (~(sp.isnan(Y).any(axis=1))==self.Iok).all(), 'VarianceDecomposition:: pattern of missing values needs to match Y given at initialization'

        # TODO: update the current code

        self.Y = Y
        self.vd.setPheno(Y)

        self.init    = False
        self.optimum = None

        self.cache['Sigma']   = None
        self.cache['Hessian'] = None

    def setTestSampleSize(self, Ntest):
        """
        set sample size of the test dataset

        Args:
            Ntest:        sample size of the test set
        """
        self.Ntest = Ntest


    def addRandomEffect(self, K=None, is_noise=False, normalize=False, Kcross=None, trait_covar_type='freeform', rank=1, fixed_trait_covar=None, jitter=1e-4):
        """
        Add random effects term.

        Args:
            K:      Sample Covariance Matrix [N, N]
            is_noise:   Boolean indicator specifying if the matrix is homoscedastic noise (weighted identity covariance) (default False)
            normalize:  Boolean indicator specifying if K has to be normalized such that K.trace()=N.
            Kcross:            NxNtest cross covariance for predictions
            trait_covar_type: type of covaraince to use. Default 'freeform'. possible values are
                            'freeform':  general semi-definite positive matrix,
                            'fixed': fixed matrix specified in fixed_trait_covar,
                            'diag': diagonal matrix,
                            'lowrank': low rank matrix. The rank of the lowrank part is specified in the variable rank,
                            'lowrank_id': sum of a lowrank matrix and a multiple of the identity. The rank of the lowrank part is specified in the variable rank,
                            'lowrank_diag': sum of a lowrank and a diagonal matrix. The rank of the lowrank part is specified in the variable rank,
                            'block': multiple of a matrix of ones,
                            'block_id': sum of a multiple of a matrix of ones and a multiple of the idenity,
                            'block_diag': sum of a multiple of a matrix of ones and a diagonal matrix,
            rank:       rank of a possible lowrank component (default 1)
            fixed_trait_covar:   PxP matrix for the (predefined) trait-to-trait covariance matrix if fixed type is used
            jitter:        diagonal contribution added to trait-to-trait covariance matrices for regularization
        """
        assert K is not None or is_noise, 'VarianceDecomposition:: Specify covariance structure'

        if is_noise:
            assert self.noisPos is None, 'VarianceDecomposition:: Noise term already exists'
            K  = sp.eye(self.N)
            Kcross = None
            self.noisPos = self.n_randEffs
        else:
            assert K.shape[0]==self.N, 'VarianceDecomposition:: Incompatible shape for K'
            assert K.shape[1]==self.N, 'VarianceDecomposition:: Incompatible shape for K'
        if Kcross is not None:
            assert self.Ntest is not None, 'VarianceDecomposition:: specify Ntest for predictions (method VarianceDecomposition::setTestSampleSize)'
            assert Kcross.shape[0]==self.N, 'VarianceDecomposition:: Incompatible shape for Kcross'
            assert Kcross.shape[1]==self.Ntest, 'VarianceDecomposition:: Incompatible shape for Kcross'

        if normalize:
            cc = covar_rescaling_factor(K) 
            K *= cc
            if Kcross is not None: Kcross *= cc 

        # add random effect
        self.sample_covars.append(K)
        self.sample_cross_covars.append(Kcross)
        _cov = self._buildTraitCovar(trait_covar_type=trait_covar_type, rank=rank, fixed_trait_covar=fixed_trait_covar, jitter=jitter)
        self.trait_covars.append(_cov)
        self.n_randEffs+=1

        #TODO: define _sync function
        self._desync()

    def addFixedEffect(self, F=None, A=None, Ftest=None):
        """
        add fixed effect term to the model

        Args:
            F:     sample design matrix for the fixed effect [N,K]
            A:     trait design matrix for the fixed effect (e.g. sp.ones((1,P)) common effect; sp.eye(P) any effect) [L,P]
            Ftest: sample design matrix for test samples [Ntest,K]
        """
        if A is None:
            A = sp.eye(self.P)
        if F is None:
            F = sp.ones((self.N,1))
            if self.Ntest is not None:
                Ftest = sp.ones((self.Ntest,1))

        assert A.shape[1]==self.P, 'VarianceDecomposition:: A has incompatible shape'
        assert F.shape[0]==self.N, 'VarianceDecimposition:: F has incompatible shape'

        if Ftest is not None:
            assert self.Ntest is not None, 'VarianceDecomposition:: specify Ntest for predictions (method VarianceDecomposition::setTestSampleSize)'
            assert Ftest.shape[0]==self.Ntest, 'VarianceDecimposition:: Ftest has incompatible shape'
            assert Ftest.shape[1]==F.shape[1], 'VarianceDecimposition:: Ftest has incompatible shape'

        # add fixed effect
        self.sample_designs.append(F)
        self.sample_test_designs.append(Ftest)
        self.trait_designs.append(A)
 
        self._desync()


    def optimize(self, init_method='default', inference=None, n_times=10, perturb=False, pertSize=1e-3, verbose=None):
        """
        Train the model using the specified initialization strategy

        Args:
            init_method:    initialization strategy:
                                'default': variance is equally split across the different random effect terms. For mulit-trait models the empirical covariance between traits is used
                                'random': variance component parameters (scales) are sampled from a normal distribution with mean 0 and std 1,
                                None: no initialization is considered. Initial parameters can be specifies by using the single covariance getTraitCovarfun() 
            inference:      inference gp method, by default algebrically efficient inference (i.e., gp2kronSum, gp2KronSumLR, gp3KronSumLR) will be used when possible.
                            For models with high a standard inference scheme (gp_base) will be used. 
            n_times:        number of restarts to converge
            perturb:        if true, the initial point (if random initializaiton is not being used) is perturbed with gaussian noise for each restart (default, False)
            perturbSize:    std of the gassian noise used to perturb the initial point
            verbose:        print if convergence is achieved and how many restarted were needed
        """
        #verbose = dlimix.getVerbose(verbose)
        assert init_method in ['default', 'random', None], 'VarianceDecomposition: specified init_method not valid'
        if init_method=='default':  self._init_params_default()

        # determine if repeats should be considered 
        if init_method is not 'random' and not perturb:     n_times = 1

        # determine inference
        if inference is None:   inference = self._det_inference()
        else:                   self._check_inference(inference)
        self._inference = inference

        if self.gp is None:        self._initGP()

        params0 = self.gp.getParams()
        for i in range(n_times):
            if init_method=='random':   
                params = {'covar': sp.randn(params0['covar'].shape[0])}
                self.gp.setParams(params)
            elif perturb:
                params = {'covar': params0['covar'] + pertSize * sp.randn(params0['covar'].shape[0])}
                self.gp.setParams(params)
            conv, info = self.gp.optimize()
            if conv:    break

        if verbose:
            if conv==False:
                print('No local minimum found for the tested initialization points')
            else:
                print(('Local minimum found at iteration %d' % i))

        return conv

    def getLML(self):
        """
        Return log marginal likelihood
        """
        assert self.init, 'VarianceDecomposition:: GP not initialised'
        return self.gp.LML()


    def getLMLgrad(self):
        """
        Return gradient of log-marginal likelihood
        """
        assert self.init, 'VarianceDecomposition:: GP not initialised'
        return self.gp.LML_grad()['covar']


    def getWeights(self, term_i=None):
        """
        Return weights for fixed effect term term_i

        Args:
            term_i:     fixed effect term index
        Returns:
            weights of the spefied fixed effect term.
            The output will be a KxL matrix of weights will be returned,
            where K is F.shape[1] and L is A.shape[1] of the correspoding fixed effect term
            (L will be always 1 for single-trait analysis). 
        """
        assert self.init, 'GP not initialised'
        if term_i==None:
            if self.gp.mean.n_terms==1:
                term_i = 0
            else:
                print('VarianceDecomposition: Specify fixed effect term index')
        return self.gp.mean.B[term_i]


    def getTraitCovarFun(self, term_i):
        """
        Return the trait limix covariance function for term term_i

        Args: 
            term_i:     index of the ranodm effect to consider
        Returns:
            LIMIX cvoariance function 
        """
        assert term_i < self.n_randEffs, 'VarianceDecomposition:: specied term out of range'
        return self.trait_covars[term_i]

    def getTraitCovar(self, term_i=None):
        """
        Return the estimated trait covariance matrix for term_i (or the total if term_i is None)
            To retrieve the matrix of correlation coefficient use \see getTraitCorrCoef

        Args:
            term_i:     index of the random effect term we want to retrieve the covariance matrix
        Returns:
            estimated trait covariance 
        """
        assert term_i < self.n_randEffs, 'VarianceDecomposition:: specied term out of range'

        if term_i is None:
            RV = sp.zeros((self.P,self.P))
            for term_i in range(self.n_randEffs):
                RV += self.getTraitCovarFun().K()
        else:
            assert term_i<self.n_randEffs, 'Term index non valid'
            RV = self.getTraitCovarFun(term_i).K()
        return RV


    def getTraitCorrCoef(self,term_i=None):
        """
        Return the estimated trait correlation coefficient matrix for term_i (or the total if term_i is None)
            To retrieve the trait covariance matrix use \see getTraitCovar

        Args:
            term_i:     index of the random effect term we want to retrieve the correlation coefficients
        Returns:
            estimated trait correlation coefficient matrix
        """
        cov = self.getTraitCovar(term_i)
        stds = sp.sqrt(cov.diagonal())[:,sp.newaxis]
        RV = cov / stds / stds.T
        return RV


    def getVarianceComps(self, univariance=False):
        """
        Return the estimated variance components

        Args:
            univariance:   Boolean indicator, if True variance components are normalized to sum up to 1 for each trait
        Returns:
            variance components of all random effects on all phenotypes [P, n_randEffs matrix]
        """
        RV=sp.zeros((self.P,self.n_randEffs))
        for term_i in range(self.n_randEffs):
            RV[:,term_i] = self.getTraitCovar(term_i).diagonal()
        if univariance:
            RV /= RV.sum(1)[:,sp.newaxis]
        return RV

    ##################################
    # INTERNAL METHODS
    #################################
    def _desync(self):
        self._inference = None
        self.gp = None

    @property
    def init(self):
        return self.gp is not None

    def _init_params_default(self):
        """
        Internal method for default parameter initialization
        """
        # if there are some nan -> mean impute
        Yimp = self.Y.copy()
        Inan = sp.isnan(Yimp)
        Yimp[Inan] = Yimp[~Inan].mean()
        if self.P==1:   C = sp.array([[Yimp.var()]])
        else:           C = sp.cov(Yimp.T)
        C /= float(self.n_randEffs)
        for ti in range(self.n_randEffs):
            self.getTraitCovarFun(ti).setCovariance(C)

    def _det_inference(self):
        """
        Internal method for determining the inference method
        """
        # 2 random effects with complete design -> gp2KronSum
        # TODO: add check for low-rankness, use GP3KronSumLR and GP2KronSumLR when possible
        if (self.n_randEffs==2) and (~sp.isnan(self.Y).any()):
            rv = 'GP2KronSum'
        else:
            rv = 'GP'
        return rv

    def _check_inference(self, inference):
        """
        Internal method for checking that the selected inference scheme is compatible with the specified model
        """
        if inference=='GP2KronSum':
            assert self.n_randEffs==2, 'VarianceDecomposition: for fast inference number of random effect terms must be == 2'
            assert not sp.isnan(self.Y).any(), 'VarianceDecomposition: fast inference available only for complete phenotype designs'
        # TODO: add GP3KronSumLR, GP2KronSumLR


    def _initGP(self):
        """
        Internal method for initialization of the GP inference objetct
        """
        if self._inference=='GP2KronSum': 
            signalPos = sp.where(sp.arange(self.n_randEffs)!=self.noisPos)[0][0]
            gp = GP2KronSum(Y=self.Y, F=self.sample_designs, A=self.trait_designs,
                            Cg=self.trait_covars[signalPos], Cn=self.trait_covars[self.noisPos],
                            R=self.sample_covars[signalPos])
        else:
            mean = MeanKronSum(self.Y, self.sample_designs, self.trait_designs) 
            Iok = vec(~sp.isnan(mean.Y))[:,0] 
            if Iok.all():   Iok = None
            covar = SumCov(*[KronCov(self.trait_covars[i], self.sample_covars[i], Iok=Iok) for i in range(self.n_randEffs)])
            gp = GP(covar = covar, mean = mean) 
        self.gp = gp


    def _buildTraitCovar(self, trait_covar_type='freeform', rank=1, fixed_trait_covar=None, jitter=1e-4):
        """
        Internal functions that builds the trait covariance matrix using the LIMIX framework

        Args:
            trait_covar_type: type of covaraince to use. Default 'freeform'. possible values are
            rank:                rank of a possible lowrank component (default 1)
            fixed_trait_covar:   PxP matrix for the (predefined) trait-to-trait covariance matrix if fixed type is used
            jitter:              diagonal contribution added to freeform covariance matrices for regularization
        Returns:
            LIMIX::Covariance for Trait covariance matrix
        """
        assert trait_covar_type in ['freeform', 'diag', 'lowrank', 'lowrank_id', 'lowrank_diag', 'block', 'block_id', 'block_diag', 'fixed'], 'VarianceDecomposition:: trait_covar_type not valid'

        if trait_covar_type=='freeform':
            cov = FreeFormCov(self.P, jitter=jitter)
        elif trait_covar_type=='fixed':
            assert fixed_trait_covar is not None, 'VarianceDecomposition:: set fixed_trait_covar'
            assert fixed_trait_covar.shape[0]==self.P, 'VarianceDecomposition:: Incompatible shape for fixed_trait_covar'
            assert fixed_trait_covar.shape[1]==self.P, 'VarianceDecomposition:: Incompatible shape for fixed_trait_covar'
            cov = FixedCov(fixed_trait_covar)
        elif trait_covar_type=='diag':
            cov = DiagonalCov(self.P)
        elif trait_covar_type=='lowrank':
            cov = LowRankCov(self.P, rank=rank)
        elif trait_covar_type=='lowrank_id':
            cov = SumCov(LowRankCov(self.P, rank=rank), FixedCov(sp.eye(self.P)))
        elif trait_covar_type=='lowrank_diag':
            cov = SumCov(LowRankCov(self.P, rank=rank), DiagonalCov(self.P))
        elif trait_covar_type=='block':
            cov = FixedCov(sp.ones([self.P, self.P]))
        elif trait_covar_type=='block_id':
            cov1 = FixedCov(sp.ones([self.P, self.P]))
            cov2 = FixedCov(sp.eye(self.P)) 
            cov = SumCov(cov1, cov2)
        elif trait_covar_type=='block_diag':
            cov1 = FixedCov(sp.ones([self.P, self.P]))
            cov2 = FixedCov(sp.eye(self.P)) 
            cov = SumCov(cov1, cov2)
        return cov



    #####################################
    # METHODS TO UPDATE (OR DEPRECATED)
    #####################################

    def optimize_with_repeates(self,fast=None,verbose=None,n_times=10,lambd=None,lambd_g=None,lambd_n=None):
        """
        Train the model repeadly up to a number specified by the users with random restarts and
        return a list of all relative minima that have been found. This list is sorted according to
        least likelihood. Each list term is a dictionary with keys "counter", "LML", and "scales".

        After running this function, the vc object will be set at the last iteration. Thus, if you
        wish to get the vc object of one of the repeats, then set the scales. For example:

        vc.setScales(scales=optimize_with_repeates_output[0]["scales"])

        Args:
            fast:       Boolean. if set to True initalize kronSumGP
            verbose:    Boolean. If set to True, verbose output is produced. (default True)
            n_times:    number of re-starts of the optimization. (default 10)
        """
        verbose = dlimix.getVerbose(verbose)

        if not self.init:       self._initGP(fast)

        opt_list = []

        fixed0 = sp.zeros_like(self.gp.getParams()['dataTerm'])

        # minimize n_times
        for i in range(n_times):

            scales1 = self._getScalesRand()
            fixed1  = 1e-1*sp.randn(fixed0.shape[0],fixed0.shape[1])
            conv = self.trainGP(fast=fast,scales0=scales1,fixed0=fixed1,lambd=lambd,lambd_g=lambd_g,lambd_n=lambd_n)

            if conv:
                # compare with previous minima
                temp=1
                for j in range(len(opt_list)):
                    if sp.allclose(abs(self.getScales()),abs(opt_list[j]['scales'])):
                        temp=0
                        opt_list[j]['counter']+=1
                        break
                if temp==1:
                    opt = {}
                    opt['counter'] = 1
                    opt['LML'] = self.getLML()
                    opt['scales'] = self.getScales()
                    opt_list.append(opt)


        # sort by LML
        LML = sp.array([opt_list[i]['LML'] for i in range(len(opt_list))])
        index   = LML.argsort()[::-1]
        out = []
        if verbose:
            print("\nLocal mimima\n")
            print("n_times\t\tLML")
            print("------------------------------------")

        for i in range(len(opt_list)):
            out.append(opt_list[index[i]])
            if verbose:
                print(("%d\t\t%f" % (opt_list[index[i]]['counter'], opt_list[index[i]]['LML'])))
                print("")

        return out


    """ standard errors """
    def getTraitCovarStdErrors(self,term_i):
        """
        Returns standard errors on trait covariances from term_i (for the covariance estimate \see getTraitCovar)

        Args:
            term_i:     index of the term we are interested in
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'

        if self.P==1:
            out = (2*self.getScales()[term_i])**2*self._getLaplaceCovar()[term_i,term_i]
        else:
            C = self.vd.getTerm(term_i).getTraitCovar()
            n_params = C.getNumberParams()
            par_index = 0
            for term in range(term_i-1):
                par_index += self.vd.getTerm(term_i).getNumberScales()
            Sigma1 = self._getLaplaceCovar()[par_index:(par_index+n_params),:][:,par_index:(par_index+n_params)]
            out = sp.zeros((self.P,self.P))
            for param_i in range(n_params):
                out += C.Kgrad_param(param_i)**2*Sigma1[param_i,param_i]
                for param_j in range(param_i):
                    out += 2*abs(C.Kgrad_param(param_i)*C.Kgrad_param(param_j))*Sigma1[param_i,param_j]
        out = sp.sqrt(out)
        return out

    def getVarianceCompStdErrors(self,univariance=False):
        """
        Return the standard errors on the estimated variance components (for variance component estimates \see getVarianceComps)

        Args:
            univariance:   Boolean indicator, if True variance components are normalized to sum up to 1 for each trait
        Returns:
            standard errors on variance components [P, n_randEffs matrix]
        """
        RV=sp.zeros((self.P,self.n_randEffs))
        for term_i in range(self.n_randEffs):
            #RV[:,term_i] = self.getTraitCovarStdErrors(term_i).diagonal()
            RV[:,term_i] = self.getTraitCovarStdErrors(term_i)
        var = self.getVarianceComps()
        if univariance:
            RV /= var.sum(1)[:,sp.newaxis]
        return RV

    """
    CODE FOR PREDICTIONS
    """
    def predictPhenos(self,use_fixed=None,use_random=None):
        """
        predict the conditional mean (BLUP)

        Args:
            use_fixed:        list of fixed effect indeces to use for predictions
            use_random:        list of random effect indeces to use for predictions
        Returns:
            predictions (BLUP)
        """
        assert self.noisPos is not None,      'No noise element'
        assert self.init,               'GP not initialised'
        assert self.Ntest is not None,        'VarianceDecomposition:: specify Ntest for predictions (method VarianceDecomposition::setTestSampleSize)'

        use_fixed  = list(range(self.n_fixedEffs))
        use_random = list(range(self.n_randEffs))

        KiY = self.gp.agetKEffInvYCache()

        if self.fast==False:
            KiY = KiY.reshape(self.P,self.N).T

        Ypred = sp.zeros((self.Ntest,self.P))

        # predicting from random effects
        for term_i in use_random:
            if term_i!=self.noisPos:
                Kstar = self.Kstar[term_i]
                if Kstar is None:
                    warnings.warn('warning: random effect term %d not used for predictions as it has None cross covariance'%term_i)
                    continue
                term  = sp.dot(Kstar.T,KiY)
                if self.P>1:
                    C    = self.getTraitCovar(term_i)
                    term = sp.dot(term,C)
                else:
                    term *= self.getVarianceComps()[0,term_i]
                Ypred += term

        # predicting from fixed effects
        weights = self.getWeights()
        w_i = 0
        for term_i in use_fixed:
            Fstar = self.Fstar[term_i]
            if Fstar is None:
                warnings.warn('warning: fixed effect term %d not used for predictions as it has None test sample design'%term_i)
                continue
            if self.P==1:    A = sp.eye(1)
            else:            A = self.vd.getDesign(term_i)
            Fstar = self.Fstar[term_i]
            W = weights[w_i:w_i+A.shape[0],0:1].T
            term = sp.dot(Fstar,sp.dot(W,A))
            w_i += A.shape[0]
            Ypred += term

        return Ypred


    def crossValidation(self,seed=0,n_folds=10,fullVector=True,verbose=None,D=None,**keywords):
        """
        Split the dataset in n folds, predict each fold after training the model on all the others

        Args:
            seed:        seed
            n_folds:     number of folds to train the model on
            fullVector:  Bolean indicator, if true it stops if no convergence is observed for one of the folds, otherwise goes through and returns a pheno matrix with missing values
            verbose:     if true, prints the fold that is being used for predicitons
            **keywords:  params to pass to the function optimize
        Returns:
            Matrix of phenotype predictions [N,P]
        """
        verbose = dlimix.getVerbose(verbose)

        # split samples into training and test
        sp.random.seed(seed)
        r = sp.random.permutation(self.Y.shape[0])
        nfolds = 10
        Icv = sp.floor(((sp.ones((self.Y.shape[0]))*nfolds)*r)/self.Y.shape[0])

        RV = {}
        if self.P==1:  RV['var'] = sp.zeros((nfolds,self.n_randEffs))
        else:          RV['var'] = sp.zeros((nfolds,self.P,self.n_randEffs))

        Ystar = sp.zeros_like(self.Y)

        for fold_j in range(n_folds):

            if verbose:
                print((".. predict fold %d"%fold_j))

            Itrain  = Icv!=fold_j
            Itest   = Icv==fold_j
            Ytrain  = self.Y[Itrain,:]
            Ytest   = self.Y[Itest,:]
            vc = VarianceDecomposition(Ytrain)
            vc.setTestSampleSize(Itest.sum())
            for term_i in range(self.n_fixedEffs):
                F      = self.vd.getFixed(term_i)
                Ftest  = F[Itest,:]
                Ftrain = F[Itrain,:]
                if self.P>1:    A = self.vd.getDesign(term_i)
                else:           A = None
                vc.addFixedEffect(F=Ftrain,Ftest=Ftest,A=A)
            for term_i in range(self.n_randEffs):
                if self.P>1:
                    tct  = self.trait_covar_type[term_i]
                    rank = self.rank[term_i]
                    ftc  = self.fixed_tc[term_i]
                    jitt = self.jitter[term_i]
                    if tct=='lowrank_diag1' or tct=='freeform1':
                        d = D[fold_j,:,term_i]
                    else:
                        d = None
                else:
                    tct  = None
                    rank = None
                    ftc  = None
                    jitt = None
                    d    = None
                if term_i==self.noisPos:
                    vc.addRandomEffect(is_noise=True,trait_covar_type=tct,rank=rank,jitter=jitt,fixed_trait_covar=ftc,d=d)
                else:
                    R = self.vd.getTerm(term_i).getK()
                    Rtrain = R[Itrain,:][:,Itrain]
                    Rcross = R[Itrain,:][:,Itest]
                    vc.addRandomEffect(K=Rtrain,Kcross=Rcross,trait_covar_type=tct,rank=rank,jitter=jitt,fixed_trait_covar=ftc,d=d)
            conv = vc.optimize(verbose=False,**keywords)
            if self.P==1:
                RV['var'][fold_j,:] = vc.getVarianceComps()[0,:]
            else:
                RV['var'][fold_j,:,:] = vc.getVarianceComps()

            if fullVector:
                assert conv, 'VarianceDecompositon:: not converged for fold %d. Stopped here' % fold_j
            if conv:
                Ystar[Itest,:] = vc.predictPhenos()
            else:
                warnings.warn('not converged for fold %d' % fold_j)
                Ystar[Itest,:] = sp.nan

        return Ystar,RV


    """ INTERNAL FUNCIONS FOR PARAMETER INITIALIZATION """

    def _getH2singleTrait(self, K, verbose=None):
        """
        Internal function for parameter initialization
        estimate variance components and fixed effect using a linear mixed model with an intercept and 2 random effects (one is noise)
        Args:
            K:        covariance matrix of the non-noise random effect term
        """
        verbose = dlimix.getVerbose(verbose)
        # Fit single trait model
        varg  = sp.zeros(self.P)
        varn  = sp.zeros(self.P)
        fixed = sp.zeros((1,self.P))

        for p in range(self.P):
            y = self.Y[:,p:p+1]
            # check if some sull value
            I  = sp.isnan(y[:,0])
            if I.sum()>0:
                y  = y[~I,:]
                _K = K[~I,:][:,~I]
            else:
                _K  = copy.copy(K)
            lmm = dlimix.CLMM()
            lmm.setK(_K)
            lmm.setSNPs(sp.ones((y.shape[0],1)))
            lmm.setPheno(y)
            lmm.setCovs(sp.zeros((y.shape[0],1)))
            lmm.setVarcompApprox0(-20, 20, 1000)
            lmm.process()
            delta = sp.exp(lmm.getLdelta0()[0,0])
            Vtot  = sp.exp(lmm.getLSigma()[0,0])

            varg[p] = Vtot
            varn[p] = delta*Vtot
            fixed[:,p] = lmm.getBetaSNP()

            if verbose: print(p)

        sth = {}
        sth['varg']  = varg
        sth['varn']  = varn
        sth['fixed'] = fixed

        return sth

    def _getScalesDiag(self,termx=0):
        """
        Internal function for parameter initialization
        Uses 2 term single trait model to get covar params for initialization

        Args:
            termx:      non-noise term terms that is used for initialization
        """
        assert self.P>1, 'VarianceDecomposition:: diagonal init_method allowed only for multi trait models'
        assert self.noisPos is not None, 'VarianceDecomposition:: noise term has to be set'
        assert termx<self.n_randEffs-1, 'VarianceDecomposition:: termx>=n_randEffs-1'
        assert self.trait_covar_type[self.noisPos] not in ['lowrank','block','fixed'], 'VarianceDecomposition:: diagonal initializaiton not posible for such a parametrization'
        assert self.trait_covar_type[termx] not in ['lowrank','block','fixed'], 'VarianceDecimposition:: diagonal initializaiton not posible for such a parametrization'
        scales = []
        res = self._getH2singleTrait(self.vd.getTerm(termx).getK())
        scaleg = sp.sqrt(res['varg'].mean())
        scalen = sp.sqrt(res['varn'].mean())
        for term_i in range(self.n_randEffs):
            if term_i==termx:
                _scales = scaleg*self.diag[term_i]
            elif term_i==self.noisPos:
                _scales = scalen*self.diag[term_i]
            else:
                _scales = 0.*self.diag[term_i]
            if self.jitter[term_i]>0:
                _scales = sp.concatenate((_scales,sp.array([sp.sqrt(self.jitter[term_i])])))
            scales.append(_scales)
        return sp.concatenate(scales)

    def _getScalesPairwise(self,verbose=False, initDiagonal=False):
        """
        Internal function for parameter initialization
        Uses a single trait model for initializing variances and
        a pairwise model to initialize correlations
        """
        var = sp.zeros((self.P,2))

        if initDiagonal:
            #1. fit single trait model
            if verbose:
                print('.. fit single-trait model for initialization')
            vc = VarianceDecomposition(self.Y[:,0:1])
            for term_i in range(self.n_randEffs):
                if term_i==self.noisPos:
                    vc.addRandomEffect(is_noise=True)
                else:
                    K = self.vd.getTerm(term_i).getK()
                    vc.addRandomEffect(K=K)
            scales0 = sp.sqrt(0.5)*sp.ones(2)

            for p in range(self.P):
                if verbose: print(('   .. trait %d' % p))
                vc.setY(self.Y[:,p:p+1])
                conv = vc.optimize(scales0=scales0)
                if not conv:
                    print('warning initialization not converged')
                var[p,:] = vc.getVarianceComps()[0,:]

        elif fastlmm_present:
            if verbose:
                print('.. fit single-trait model for initialization (using fastlmm)')
            for p in range(self.P):
                if verbose: print(('   .. trait %d' % p))
                covariates = None
                for term_i in range(self.n_randEffs):
                    if term_i==self.noisPos:
                        pass
                    else:
                        K = self.vd.getTerm(term_i).getK()
                varY = sp.var(self.Y[:,p:p+1])
                lmm = fastLMM(X=covariates, Y=self.Y[:,p:p+1], G=None, K=K)
                opt = lmm.findH2(nGridH2=100)
                h2 = opt['h2']
                var[p,:] = h2 * varY
                var[p,self.noisPos] = (1.0-h2) * varY
                #;ipdb.set_trace()
        else:
            if verbose:
                print('.. random initialization of diagonal')
            var = sp.random.randn(var.shape[0],var.shape[1])
            var = var*var + 0.001
        #2. fit pairwise model
        if verbose:
            print('.. fit pairwise model for initialization')
        vc = VarianceDecomposition(self.Y[:,0:2])
        for term_i in range(self.n_randEffs):
            if term_i==self.noisPos:
                vc.addRandomEffect(is_noise=True,trait_covar_type='freeform')
            else:
                K = self.vd.getTerm(term_i).getK()
                vc.addRandomEffect(K=K,trait_covar_type='freeform')
        rho_g = sp.ones((self.P,self.P))
        rho_n = sp.ones((self.P,self.P))
        for p1 in range(self.P):
            for p2 in range(p1):
                if verbose:
                    print(('   .. fit pair (%d,%d)'%(p1,p2)))
                vc.setY(self.Y[:,[p1,p2]])
                scales0 = sp.sqrt(sp.array([var[p1,0],1e-4,var[p2,0],1e-4,var[p1,1],1e-4,var[p2,1],1e-4]))
                conv = vc.optimize(scales0=scales0)
                if not conv:
                    print('warning initialization not converged')
                Cg = vc.getTraitCovar(0)
                Cn = vc.getTraitCovar(1)
                rho_g[p1,p2] = Cg[0,1]/sp.sqrt(Cg.diagonal().prod())
                rho_n[p1,p2] = Cn[0,1]/sp.sqrt(Cn.diagonal().prod())
                rho_g[p2,p1] = rho_g[p1,p2]
                rho_n[p2,p1] = rho_n[p1,p2]
        #3. init
        Cg0 = rho_g*sp.dot(sp.sqrt(var[:,0:1]),sp.sqrt(var[:,0:1].T))
        Cn0 = rho_n*sp.dot(sp.sqrt(var[:,1:2]),sp.sqrt(var[:,1:2].T))
        offset_g = abs(sp.minimum(sp.linalg.eigh(Cg0)[0].min(),0))+1e-4
        offset_n = abs(sp.minimum(sp.linalg.eigh(Cn0)[0].min(),0))+1e-4
        Cg0+=offset_g*sp.eye(self.P)
        Cn0+=offset_n*sp.eye(self.P)
        Lg = sp.linalg.cholesky(Cg0)
        Ln = sp.linalg.cholesky(Cn0)
        Cg_params0 = sp.concatenate([Lg[:,p][:p+1] for p in range(self.P)])
        Cn_params0 = sp.concatenate([Ln[:,p][:p+1] for p in range(self.P)])
        scales0 = sp.concatenate([Cg_params0,1e-2*sp.ones(1),Cn_params0,1e-2*sp.ones(1)])

        return scales0


    def _getScalesRand(self):
        """
        Internal function for parameter initialization
        Return a vector of random scales
        """
        if self.P>1:
            scales = []
            for term_i in range(self.n_randEffs):
                _scales = sp.randn(self.diag[term_i].shape[0])
                if self.jitter[term_i]>0:
                    _scales = sp.concatenate((_scales,sp.array([sp.sqrt(self.jitter[term_i])])))
                scales.append(_scales)
            scales = sp.concatenate(scales)
        else:
            scales=sp.randn(self.vd.getNumberScales())
        return scales

    def _perturbation(self):
        """
        Internal function for parameter initialization
        Returns Gaussian perturbation
        """
        if self.P>1:
            scales = []
            for term_i in range(self.n_randEffs):
                _scales = sp.randn(self.diag[term_i].shape[0])
                if self.jitter[term_i]>0:
                    _scales  = sp.concatenate((_scales,sp.zeros(1)))
                scales.append(_scales)
            scales = sp.concatenate(scales)
        else:
            scales = sp.randn(self.vd.getNumberScales())
        return scales

    """ INTERNAL FUNCIONS FOR ESTIMATING PARAMETERS UNCERTAINTY """

    def _getHessian(self):
        """
        Internal function for estimating parameter uncertainty
        COMPUTES OF HESSIAN OF E(\theta) = - log L(\theta | X, y)
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'

        if self.cache['Hessian'] is None:
            ParamMask=self.gp.getParamMask()['covar']
            std=sp.zeros(ParamMask.sum())
            H=self.gp.LMLhess_covar()
            It= (ParamMask[:,0]==1)
            self.cache['Hessian']=H[It,:][:,It]

        return self.cache['Hessian']


    def _getLaplaceCovar(self):
        """
        Internal function for estimating parameter uncertainty
        Returns:
            the
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'

        if self.cache['Sigma'] is None:
            self.cache['Sigma'] = sp.linalg.inv(self._getHessian())
        return self.cache['Sigma']

    def _getFisher(self):
        """
        Return the fisher information matrix over the parameters
        TO CHECK: IT DOES NOT WORK PROPERLY
        """
        Ctot = self.vd.getGP().getCovar()
        Ki = sp.linalg.inv(Ctot.K())
        n_scales = self.vd.getNumberScales()
        out = sp.zeros((n_scales,n_scales))
        for m in range(n_scales):
            out[m,m] = 0.5 * sp.trace(sp.dot(Ki,sp.dot(Ctot.Kgrad_param(m),sp.dot(Ki,Ctot.Kgrad_param(m)))))
            for n in range(m):
                out[m,n]=0.5 * sp.trace(sp.dot(Ki,sp.dot(Ctot.Kgrad_param(m),sp.dot(Ki,Ctot.Kgrad_param(n)))))
                out[n,m]=out[m,n]
        return out

    """ DEPRECATED """

    def _getModelPosterior(self,min):
        """
        USES LAPLACE APPROXIMATION TO CALCULATE THE BAYESIAN MODEL POSTERIOR
        """
        Sigma = self._getLaplaceCovar(min)
        n_params = self.vd.getNumberScales()
        ModCompl = 0.5*n_params*sp.log(2*sp.pi)+0.5*sp.log(sp.linalg.det(Sigma))
        RV = min['LML']+ModCompl
        return RV

    def trainGP(self,fast=None,scales0=None,fixed0=None,lambd=None,lambd_g=None,lambd_n=None):
        print('DEPRECATED')
    



import os
import sys
from os.path import join

from Cython.Distutils import build_ext
from setuptools import find_packages, setup
from setuptools.extension import Extension

try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except (OSError, IOError, ImportError):
    long_description = open('README.md').read()


def _check_gcc_cpp11(cc_name):
    import subprocess
    try:
        cmd = cc_name + ' -E -dM -std=c++11 -x c++ /dev/null > /dev/null'
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError:
        return False
    return True


class build_ext_subclass(build_ext):
    def build_extensions(self):
        import platform
        from distutils import sysconfig
        if (hasattr(self.compiler, 'compiler') and
                len(self.compiler.compiler) > 0):
            cc_name = self.compiler.compiler[0]
            stdcpp = '-std=c++11'
            if 'gcc' in cc_name and not _check_gcc_cpp11(cc_name):
                stdcpp = '-std=c++0x'
            for e in self.extensions:
                e.extra_compile_args.append(stdcpp)

            conf_vars = sysconfig.get_config_vars()
            if 'MACOSX_DEPLOYMENT_TARGET' in conf_vars and len(conf_vars[
                    'MACOSX_DEPLOYMENT_TARGET']) > 0:
                _v1, _v2 = conf_vars['MACOSX_DEPLOYMENT_TARGET'].split('.')
                if int(_v1) == 10 and int(_v2) < 9:
                    stdcpp = '--stdlib=libc++'
                    for e in self.extensions:
                        e.extra_compile_args.append(stdcpp)
        build_ext.build_extensions(self)


def mac_workaround(compatible):
    import platform
    from distutils import sysconfig

    conf_vars = sysconfig.get_config_vars()
    if compatible:
        conf_vars['MACOSX_DEPLOYMENT_TARGET'] = '10.9'
        return
    vers = platform.mac_ver()[0].split('.')
    if len(vers) == 3:
        conf_vars['MACOSX_DEPLOYMENT_TARGET'] =\
            vers[0] + '.' + vers[1]
    else:
        conf_vars['MACOSX_DEPLOYMENT_TARGET'] = platform.mac_ver()[0]


def extra_compile_args():
    if sys.platform.startswith('win'):
        return []
    return [
        '-Wno-comment', '-Wno-unused-but-set-variable',
        '-Wno-overloaded-virtual', '-Wno-uninitialized',
        '-Wno-delete-non-virtual-dtor', '-Wunused-variable'
    ]


def core_extension(reswig):
    import numpy as np

    def globr(root, pattern):
        import fnmatch

        matches = []
        for root, _, filenames in os.walk(root):
            for filename in fnmatch.filter(filenames, pattern):
                matches.append(os.path.join(root, filename))

        return matches

    def swig_opts():
        return [
            '-c++', '-outdir', join('limix', 'deprecated'), '-I' + join('src')
        ]

    def nlopt_files():
        src = open(join('External', 'nlopt_src.files')).readlines()
        src = [join('External', 'nlopt', s).strip() for s in src]
        hdr = globr(join('External', 'nlopt'), '*/*.h')
        return (src, hdr)

    (src, hdr) = nlopt_files()
    src.extend(globr(join('src', 'limix'), '*.cpp'))
    hdr.extend(globr(join('src', 'limix'), '*.h'))

    incl = ['src', 'External', join('External', 'nlopt')]
    incl = [join(i) for i in incl]
    folder = join('External', 'nlopt')
    incl += [join(folder, f) for f in os.listdir(folder)]
    incl = [i for i in incl if os.path.isdir(i)]
    incl.extend([np.get_include()])

    wrap_file = join('src', 'interfaces', 'python', 'limix_wrap.cpp')
    i_file = join('src', 'interfaces', 'python', 'limix.i')

    if os.path.exists(wrap_file):
        src.append(wrap_file)
    else:
        src.append(i_file)

    depends = src + hdr

    ext = Extension(
        'limix.deprecated._core',
        src,
        include_dirs=incl,
        extra_compile_args=extra_compile_args(),
        swig_opts=swig_opts(),
        depends=depends)

    return ext


def ensemble_extension():
    import numpy as np

    src = [join('cython', 'lmm_forest', 'SplittingCore.pyx')]
    incl = [join('External'), np.get_include()]
    depends = src
    ext = Extension(
        'limix.ensemble.SplittingCore',
        src,
        language='c++',
        include_dirs=incl,
        extra_compile_args=extra_compile_args(),
        depends=depends)
    from Cython.Build import cythonize
    return cythonize(ext)


def setup_package(reswig, compatible):
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    if sys.platform == 'darwin':
        mac_workaround(compatible)

    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    pytest_runner = ['pytest-runner>=2.9'] if needs_pytest else []

    install_requires = [
        "scikit-learn", "pandas", "hcache", "scipy", "h5py", "matplotlib"
    ]
    setup_requires = [] + pytest_runner
    tests_require = ['pytest']

    metadata = dict(
        name="limix",
        version="0.8.1",
        description="A flexible and fast mixed model toolbox.",
        long_description=long_description,
        keywords='linear mixed models, GWAS, QTL, ' +
        'Variance component modelling',
        maintainer="Limix Developers",
        author="Christoph Lippert, Paolo Casale, Oliver Stegle",
        author_email="stegle@ebi.ac.uk",
        maintainer_email="stegle@ebi.ac.uk",
        packages=find_packages(),
        include_package_data=True,
        url='http://pmbio.github.io/limix/',
        install_requires=install_requires,
        setup_requires=setup_requires,
        tests_require=tests_require,
        zip_safe=False,
        license='Apache License 2.0',
        ext_modules=[core_extension(reswig)] + ensemble_extension(),
        cmdclass=dict(build_ext=build_ext_subclass),
        entry_points={
            'console_scripts': [
                'limix_runner=limix.scripts.limix_runner:entry_point',
                'mtSet_postprocess=limix.scripts.mtSet_postprocess:entry_point',
                'mtSet_preprocess=limix.scripts.mtSet_preprocess:entry_point',
                'mtSet_simPheno=limix.scripts.mtSet_simPheno:entry_point',
                'mtSet_analyze=limix.scripts.mtSet_analyze:entry_point',
                'limix_converter=limix.scripts.limix_converter:entry_point',
                'iSet_analyze=limix.scripts.iSet_analyze:entry_point',
                'iSet_postprocess=limix.scripts.iSet_postprocess:entry_point',
            ]
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Natural Language :: English',
            'Programming Language :: Python :: 2.7',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ], )

    # http://stackoverflow.com/a/29634231
    import distutils.sysconfig
    cfg_vars = distutils.sysconfig.get_config_vars()
    for key, value in list(cfg_vars.items()):
        if type(value) == str:
            cfg_vars[key] = value.replace("-Wstrict-prototypes", "")

    try:
        setup(**metadata)
    finally:
        del sys.path[0]
        os.chdir(old_path)


if __name__ == '__main__':
    reswig = False
    if "--reswig" in sys.argv:
        reswig = True
        sys.argv.remove("--reswig")

    compatible = False
    if "--compatible" in sys.argv:
        compatible = True
        sys.argv.remove("--compatible")

    setup_package(reswig, compatible)



import scipy as SP
import pdb
SP.random.seed(0)

if __name__=='__main__':

    N = 500
    P = 4
    S = 1000
    R = 10

    hr = 0.05
    hg = 0.40

    Xr = 1.*(SP.rand(N,R)<0.2)
    Xg = 1.*(SP.rand(N,S)<0.2)
    Xg-= Xg.mean(0)
    Xg/= Xg.std(0)
    XX = SP.dot(Xg,Xg.T)
    XX/= XX.diagonal().mean()
    XX+= 1e-4*SP.eye(XX.shape[0])

    Yr = SP.dot(Xr,SP.randn(R,P))
    Yr*= SP.sqrt(hr/Yr.var(0))
    Yg = SP.dot(Xg,SP.randn(S,P))
    Yg*= SP.sqrt(hg/Yg.var(0))
    Yn = SP.randn(N,P)
    Yn*= SP.sqrt((1-hr-hg)/Yn.var(0))

    Y = Yr+Yg+Yn
    Y-= Y.mean(0)
    Y/= Y.std(0)

    SP.savetxt('Xr.txt',Xr,fmt='%d')
    SP.savetxt('XX.txt',XX,fmt='%.6f')
    SP.savetxt('Y.txt',Y,fmt='%.6f')




"""LMM testing code"""
import unittest
import scipy as SP
import numpy as np
import sys
from limix.core.covar import FreeFormCov
from limix.utils.check_grad import mcheck_grad

class TestFreeForm(unittest.TestCase):
    def setUp(self):
        SP.random.seed(1)
        self.n=4
        self.C = FreeFormCov(self.n)
        self.name = 'freeform'
        self.n_params=self.C.getNumberParams()
        params=SP.randn(self.n_params)
        self.C.setParams(params)

    def test_grad(self):
        def func(x, i):
            self.C.setParams(x)
            return self.C.K()

        def grad(x, i):
            self.C.setParams(x)
            return self.C.K_grad_i(i)

        x0 = self.C.getParams()
        err = mcheck_grad(func, grad, x0)

        np.testing.assert_almost_equal(err, 0., decimal=6)

    def test_param_activation(self):
        self.assertEqual(len(self.C.getParams()), 10)
        self.C.act_K = False
        self.assertEqual(len(self.C.getParams()), 0)

        self.C.setParams(np.array([]))
        with self.assertRaises(ValueError):
            self.C.setParams(np.array([0]))

        with self.assertRaises(ValueError):
            self.C.K_grad_i(0)

    def test_Khess(self):

        cov = self.C

        for j in range(cov.getNumberParams()):

            def func(x, i):
                cov.setParams(x)
                return cov.K_grad_i(j)

            def grad(x, i):
                cov.setParams(x)
                return cov.K_hess_i_j(j, i)

            x0 = cov.getParams()
            err = mcheck_grad(func, grad, x0)
            np.testing.assert_almost_equal(err, 0.)

if __name__ == '__main__':
    unittest.main()



import sys
import limix
from limix.core.covar import LowRankCov
from limix.core.covar import FixedCov
from limix.core.covar import FreeFormCov
from limix.core.covar import CategoricalLR
from limix.core.mean import MeanBase
from limix.core.gp import GP 
import scipy as sp
import scipy.stats as st
from limix.mtSet.core.iset_utils import * 
import numpy as np
import numpy.linalg as nla
import scipy.linalg as la
import copy
import pdb
from limix.utils.preprocess import gaussianize
from scipy.optimize import fmin
import time
import pandas as pd
from .linalg_utils import msqrt
from .linalg_utils import lowrank_approx

ntype_dict = {'assoc':'null', 'gxe':'block', 'gxehet':'rank1'}

def define_gp(Y, Xr, mean, Ie, type):
    P = 2
    if type=='null':
        _Cr = FixedCov(sp.ones([2, 2]))
        _Cr.scale = 1e-9
        _Cr.act_scale = False
        covar = CategoricalLR(_Cr, sp.ones((Xr.shape[0], 1)), Ie)
    else:
        if type=='block':         _Cr = FixedCov(sp.ones((P,P)))
        elif type=='rank1':         _Cr = LowRankCov(P,1)
        elif type=='full':          _Cr = FreeFormCov(P)
        else:                       print('poppo')
        covar = CategoricalLR(_Cr, Xr, Ie)
    _gp = GP(covar=covar, mean=mean)
    return _gp

class MvSetTestInc():

    def __init__(self, Y=None, Xr=None, F=None, factr=1e7, Ie=None, debug=False):
        """
        Args:
            Y:          [N, 1] phenotype matrix
            Xr:         [N, S] genotype data of the set component
            R:          [N, S] genotype data of the set component
            factr:      paramenter that determines the accuracy of the solution
                        (see scipy.optimize.fmin_l_bfgs_b for more details)
        """
        if F is None:
            F = sp.ones((y.shape[0], 1))
        # kroneckerize F
        W = sp.zeros((Y.shape[0], 2*F.shape[1]))
        W[:, :F.shape[1]] = Ie[:, sp.newaxis] * F 
        W[:, F.shape[1]:] = (~Ie[:, sp.newaxis]) * F 
        self.mean = MeanBase(Y, W)
        # avoid SVD failus by adding some jitter
        Xr+= 2e-6*(sp.rand(*Xr.shape)-0.5)
        # store stuff 
        Xr-= Xr.mean(0)
        Xr/= Xr.std(0)
        Xr/= sp.sqrt(Xr.shape[1])
        self.Y = Y
        self.F = F
        self.Xr = Xr
        self.Ie  = Ie
        self.covY = sp.cov(Y.T)
        self.factr = factr 
        self.debug = debug
        self.gp = {}
        self.info = {}

    def assoc(self):
        # fit model 
        for key in ['null', 'full']:
            if key not in list(self.gp.keys()):
                if self.debug:      print('.. dening %s' % key)
                self.gp[key] = define_gp(self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:      print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['null']['LML']-self.info['full']['LML']

    def gxe(self):
        # fit model 
        for key in ['null', 'full', 'block']:
            if key not in list(self.gp.keys()):
                if self.debug:      print('.. defining %s' % key)
                self.gp[key] = define_gp(self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:      print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['block']['LML']-self.info['full']['LML']

    def gxehet(self):
        # fit model
        for key in ['null', 'full', 'rank1']:
            if key not in list(self.gp.keys()):
                if self.debug:      print('.. defining %s' % key)
                self.gp[key] = define_gp(self.Y, self.Xr, self.mean, self.Ie, key)
                if self.debug:      print('.. fitting %s' % key)
                self.info[key] = self._fit(key, vc=True)
        return self.info['rank1']['LML']-self.info['full']['LML']

    def assoc_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            idx_perms = sp.random.permutation(self.Y.shape[0])
            _Xr = self.Xr[idx_perms]
            mvset0 = MvSetTestInc(Y=self.Y, F=self.F, Xr=_Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.assoc()
        return LLR0

    def gxe_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            Xb = sp.dot(self.mean.W, self.mean.b)
            _Y = Xb+self.gp['block'].covar.Kh_dot(sp.randn(self.Y.shape[0],1))
            mvset0 = MvSetTestInc(Y=_Y, F=self.F, Xr=self.Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.gxe()
        return LLR0

    def gxehet_null(self, n_nulls=30):
        LLR0 = sp.zeros(n_nulls)
        for ni in range(n_nulls):
            Xb = sp.dot(self.mean.W, self.mean.b)
            _Y = Xb+self.gp['rank1'].covar.Kh_dot(sp.randn(self.Y.shape[0],1))
            mvset0 = MvSetTestInc(Y=_Y, F=self.F, Xr=self.Xr, Ie=self.Ie)
            LLR0[ni] = mvset0.gxehet()
        return LLR0

    def _fit(self, type, vc=False):
        #2. init
        if type=='null':
            self.gp[type].covar.Cn.setCovariance(sp.eye(2))
        elif type=='full':
            Cr0_K = 1e-4*sp.ones((2,2))+1e-4*sp.eye(2)
            Cn0_K = 0.99*self.gp['null'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Cr0_K)
            self.gp[type].covar.Cn.setCovariance(Cn0_K)
        elif type=='block':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.scale = sp.mean(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        elif type=='rank1':
            Crf_K = self.gp['full'].covar.Cr.K()
            Cnf_K = self.gp['full'].covar.Cn.K()
            self.gp[type].covar.Cr.setCovariance(Crf_K)
            self.gp[type].covar.Cn.setCovariance(Cnf_K)
        else:
            print('poppo')
        conv = self.gp[type].optimize(factr=self.factr, verbose=False)[0]
        B = self.gp[type].mean.b.reshape((self.mean.W.shape[1]/2,2), order='F')
        RV = {'Cr': self.gp[type].covar.Cr.K(),      
                'Cn': self.gp[type].covar.Cn.K(),
                'B': B,
                'conv': sp.array([conv]),
                'LML': sp.array([self.gp[type].LML()]),
                'LMLgrad': sp.array([sp.mean((self.gp[type].LML_grad()['covar'])**2)])}
        if vc:
            # tr(P WW) = tr(PWWP) = ((PW)**2).sum()
            # tr(P D) = (PD).sum() = D.sum() - 1/n * (Ones*D).sum()
            #                      = D.sum() - D.sum() 
            PW = self.gp[type].covar.W()
            PW-= PW.mean(0) 
            var_r = (PW**2).sum()/ float(self.Y.size-1)
            var_c = sp.var(sp.dot(self.mean.W, self.gp[type].mean.b))
            D = self.gp[type].covar.d_inv()**(-1)
            var_n = (1-1/float(D.shape[0]))*D.sum()/float(self.Y.size-1) 
            #var_n = sp.diagonal(sp.diag(D)-sp.diag(D).mean(0)).sum()/float(self.Y.size-1)
            RV['var'] = sp.array([var_r, var_c, var_n])
            if 0 and self.Y.size<5000:
                pdb.set_trace()
                Kr = sp.kron(RV['Cr'], sp.dot(self.Xr, self.Xr.T))
                Kn = sp.kron(RV['Cn'], sp.eye(self.Y.shape[0]))
                _var_r = sp.trace(Kr-Kr.mean(0)) / float(self.Y.size-1)
                _var_n = sp.trace(Kn-Kn.mean(0)) / float(self.Y.size-1)
                _var = sp.array([_var_r, var_c, _var_n])
                print(((_var-RV['var'])**2).mean())
            if type=='full':
                trRr = (self.Xr**2).sum()
                # calculate within region vcs 
                Cr_block = sp.mean(RV['Cr']) * sp.ones(RV['Cr'].shape)
                Cr_rank1 = lowrank_approx(RV['Cr'], rank=1)
                var_block = sp.trace(Cr_block)*trRr / float(self.Y.size-1)
                var_rank1 = sp.trace(Cr_rank1)*trRr / float(self.Y.size-1)
                RV['var_r'] = sp.array([var_block, var_rank1-var_block, var_r-var_rank1])
        return RV

if 0:
    def _sim_from(self, set_covar='block', seed=None, qq=False):
        ##1. region term
        if set_covar=='block':
            Cr = self.block['Cr']
            Cg = self.block['Cg']
            Cn = self.block['Cn']
        if set_covar=='rank1':
            Cr = self.lr['Cr']
            Cg = self.lr['Cg']
            Cn = self.lr['Cn']
        Lc = msqrt(Cr)
        U, Sh, V = nla.svd(self.Xr, full_matrices=0)
        Lr = sp.zeros((self.Y.shape[0], self.Y.shape[0]))
        Lr[:, :Sh.shape[0]] = U * Sh[sp.newaxis, :]
        Z = sp.randn(*self.Y.shape)
        Yr = sp.dot(Lr, sp.dot(Z, Lc.T))
        ##2. bg term
        Lc = msqrt(Cg)
        Lr = self.XXh
        Z = sp.randn(*self.Y.shape)
        Yg = sp.dot(Lr, sp.dot(Z, Lc.T))
        # noise terms
        Lc = msqrt(Cn)
        Z = sp.randn(*self.Y.shape)
        Yn = sp.dot(Z, Lc.T)
        # normalize
        Y = Yr + Yg + Yn
        if qq:
            Y = gaussianize(Y)
            Y-= Y.mean(0)
            Y/= Y.std(0)
        return Y

if __name__=='__main__':


    if 1:

        N = 1000
        S = 20
        Xr = 1.*(sp.rand(N,S)<0.2)
        Ie = sp.randn(N)<0.
        Y = sp.randn(N, 1)
        F = sp.ones((N,1))

        pdb.set_trace()

        t0 = time.time()
        mvset = MvSetTestInc(Y=Y, Xr=Xr, F=F, Ie=Ie, factr=1e7)
        mvset.assoc()
        mvset.gxe()
        mvset.gxehet()
        print('.. permutations')
        mvset.assoc_null()
        print('.. bootstrap gxe')
        mvset.gxe_null()
        print('.. bootstrap gxehet')
        mvset.gxehet_null()
        print(time.time()-t0)

        pdb.set_trace()




#! /usr/bin/env python
# Copyright(c) 2014, The mtSet developers (Francesco Paolo Casale, Barbara Rakitsch, Oliver Stegle)
# All rights reserved.

from optparse import OptionParser
from limix.mtSet.core.iset_utils import calc_emp_pv_eff
import pandas as pd
import glob
import os
import time
import sys

def entry_point():

    parser = OptionParser()
    parser.add_option("--resdir", dest='resdir', type=str, default='./')
    parser.add_option("--outfile", dest='outfile', type=str, default=None)
    #parser.add_option("--manhattan_plot", dest='manhattan',action="store_true",default=False)
    parser.add_option("--tol", dest='tol', type=float, default=4e-3)
    (options, args) = parser.parse_args()


    resdir = options.resdir
    out_file = options.outfile
    tol = options.tol

    print('.. load permutation results')
    file_name = os.path.join(resdir, '*.iSet.perm')
    files = glob.glob(file_name)
    df0 = pd.DataFrame()
    for _file in files:
        print(_file)
        df0 = df0.append(pd.read_csv(_file, index_col=0))

    print('.. load real results')
    file_name = os.path.join(resdir, '*.iSet.real')
    files = glob.glob(file_name)
    df = pd.DataFrame()
    for _file in files:
        print(_file)
        df = df.append(pd.read_csv(_file, index_col=0))

    #calculate P values for the three tests
    for test in ['mtSet', 'iSet', 'iSet-het']:
        df[test+' pv'] = calc_emp_pv_eff(df[test+' LLR'].values,
                                         df0[test+' LLR0'].values)

    print(('.. saving %s' % out_file+'.res'))
    df.to_csv(out_file+'.res')

    if 0:
        if options.manhattan:
            import limix.utils.plot as plot

            if not os.path.exists(options.outfile):
                os.makedirs(options.outfile)

            def plot_manhattan(pv, out_file):
                import matplotlib.pylab as PLT
                import scipy as SP
                posCum = SP.arange(pv.shape[0])
                idx=~SP.isnan(pv)
                plot.plot_manhattan(posCum[idx],pv[idx],alphaNS=1.0,alphaS=1.0)
                PLT.savefig(out_file)

            for test in ['mtSet', 'iSet', 'iSet-het']:
                out_file = os.path.join(options.outfile,
                                        'iSet.%s_pv.manhattan.png'\
                                        % (test,))
                print((".. saving " + out_file))
                plot_manhattan(df['%s pv' % test].values, out_file)



import scipy as SP
import pdb

if __name__=='__main__':

    Cg_mtSet1VCNull   = SP.loadtxt('./res_mtSet1VCnull_base_Cg.txt')
    Cn_mtSet1VCNull   = SP.loadtxt('./res_mtSet1VCnull_base_Cn.txt')

    Cg_mtSetNull   = SP.loadtxt('./res_mtSetNull_base_Cg.txt')
    Cn_mtSetNull   = SP.loadtxt('./res_mtSetNull_base_Cn.txt')
    Cg_mtSetNullec = SP.loadtxt('./res_mtSetNull_eigenCache_Cg.txt')
    Cn_mtSetNullec = SP.loadtxt('./res_mtSetNull_eigenCache_Cn.txt')

    Cr_mtSet   = SP.loadtxt('./res_mtSet_base_Cr.txt')
    Cg_mtSet   = SP.loadtxt('./res_mtSet_base_Cg.txt')
    Cn_mtSet   = SP.loadtxt('./res_mtSet_base_Cn.txt')
    Cr_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cr.txt')
    Cg_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cg.txt')
    Cn_mtSetec = SP.loadtxt('./res_mtSet_eigenCache_Cn.txt')

    # Cg_mtSet1VCNull = 0
    print(((Cg_mtSet1VCNull==0).all()))

    # Eigenvalue chaching works
    print((((Cg_mtSetNull-Cg_mtSetNullec)**2<1e-4).all()))
    print((((Cn_mtSetNull-Cn_mtSetNullec)**2<1e-4).all()))
    print((((Cr_mtSet-Cr_mtSetec)**2<1e-4).all()))
    print((((Cg_mtSet-Cg_mtSetec)**2<1e-4).all()))
    print((((Cn_mtSet-Cn_mtSetec)**2<1e-4).all()))




#! /usr/bin/env python
# Copyright(c) 2014, The mtSet developers (Francesco Paolo Casale, Barbara Rakitsch, Oliver Stegle)
# All rights reserved.

import time
import sys
import os
from limix.iSet.iset import fit_iSet
from optparse import OptionParser
import numpy as np
import pandas as pd
import scipy as sp
import csv

from ..mtSet.core.read_utils import readNullModelFile
from ..mtSet.core.read_utils import readWindowsFile
from ..mtSet.core.read_utils import readCovarianceMatrixFile
from ..mtSet.core.read_utils import readCovariatesFile
from ..mtSet.core.read_utils import readPhenoFile
from ..mtSet.core import plink_reader

def entry_point():
    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    #parser.add_option("--cfile", dest='cfile', type=str, default=None)
    parser.add_option("--pfile", dest='pfile', type=str, default=None)
    parser.add_option("--wfile", dest='wfile', type=str, default=None)
    parser.add_option("--ffile", dest='ffile', type=str, default=None)
    parser.add_option("--ifile", dest='ifile', type=str, default=None)
    parser.add_option("--resdir", dest='resdir', type=str, default='./')

    # start window, end window and permutations
    parser.add_option("--minSnps", dest='minSnps', type=int, default=4)
    parser.add_option("--n_perms", type=int, default=10)
    parser.add_option("--start_wnd", dest='i0', type=int, default=None)
    parser.add_option("--end_wnd", dest='i1', type=int, default=None)
    parser.add_option("--factr", dest='factr', type=float, default=1e7)

    (options, args) = parser.parse_args()

    print('importing data')
    F = sp.loadtxt(options.ffile+'.fe')
    Y = sp.loadtxt(options.pfile+'.phe')
    if len(Y.shape)==1: Y = Y[:,sp.newaxis]

    wnds = readWindowsFile(options.wfile)

    bim = plink_reader.readBIM(options.bfile,usecols=(0,1,2,3))
    fam = plink_reader.readFAM(options.bfile,usecols=(0,1))
    chrom = bim[:, 0].astype(float)
    pos = bim[:, -1].astype(float)

    i0 = 1 if options.i0 is None else options.i0
    i1 = wnds.shape[0] if options.i1 is None else options.i1

    df = pd.DataFrame()
    df0 = pd.DataFrame()

    if options.ifile is None:
        Ie = None
    else:
        Ie = sp.loadtxt(options.ifile+'.ind').flatten()==1

    res_dir = options.resdir

    if not os.path.exists(res_dir):
        os.makedirs(res_dir)

    n_digits = len(str(wnds.shape[0]))
    fname = str(i0).zfill(n_digits)
    fname+= '_'+str(i1).zfill(n_digits)
    resfile = os.path.join(res_dir, fname)

    for wnd_i in range(i0,i1):
        t0 = time.time()
        print(('.. window %d - (%d, %d-%d) - %d snps'%(wnd_i,int(wnds[wnd_i,1]),int(wnds[wnd_i,2]),int(wnds[wnd_i,3]),int(wnds[wnd_i,-1]))))
        Xr = plink_reader.readBED(options.bfile, useMAFencoding=True, start = int(wnds[wnd_i,4]), nSNPs = int(wnds[wnd_i,5]), bim=bim , fam=fam)['snps']
        Xr = np.ascontiguousarray(Xr)
        xr = sp.dot(sp.rand(Xr.shape[0]), Xr)
        idxs_u = sp.sort(sp.unique(xr, return_index=True)[1])
        if idxs_u.shape[0]<options.minSnps:
            print('SKIPPED: number of snps lower than minSnps')
            continue
        Xr = Xr[:,idxs_u]
        Xr-= Xr.mean(0)
        Xr/= Xr.std(0)
        Xr/= np.sqrt(Xr.shape[1])
        _df, _df0 = fit_iSet(Y, F=F, Xr=Xr, Ie=Ie, n_nulls=10)
        df  = df.append(_df)
        df0 = df0.append(_df0)
        print('Elapsed:', time.time()-t0)

    df.to_csv(resfile + '.iSet.real')
    df0.to_csv(resfile + '.iSet.perm')



# import unittest
# import numpy as np
# from limix.core.mean.mean_base import MeanBase as lin_mean
# from limix.core.covar.sqexp import SQExpCov
# from limix.core.covar.fixed import fixed
# from limix.core.covar.combinators import sumcov
# from limix.core.gp.gp_base_new import gp as gp_base
#
# class TestGP(unittest.TestCase):
#     def setUp(self):
#         np.random.seed(1)
#         N = 400
#         s_x = 0.05
#         s_y = 0.1
#         X = (np.linspace(0,2,N)+s_x*np.random.randn(N))[:,np.newaxis]
#         Y = np.sin(X)+s_y*np.random.randn(N,1)
#         #pl.plot(x,y,'x')
#
#         # define mean term
#         F = 1.*(np.random.rand(N,2)<0.2)
#         mean = lin_mean(Y,F)
#
#         # define covariance matrices
#         covar1 = sqexp(X)
#         covar2 = FixedCov(np.eye(N))
#         covar  = sumcov(covar1,covar2)
#
#         # self._gp = gp_base(covar=covar,mean=mean)
#         # covar.setRandomParams()
#
#     def test_lml(self):
#         return
#         # self._gp.LML()
#
#     def test_lml_grad(self):
#         return
#         # self._gp.LML_grad()
#
# if __name__ == '__main__':
#     unittest.main()



#! /usr/bin/env python
# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
# All rights reserved.
#
# LIMIX is provided under a 2-clause BSD license.
# See LICENSE for the complete license.

import subprocess, sys, os
import time
import h5py
import scipy as sp
#import limix
from optparse import OptionParser
import pandas as pd
import limix.deprecated.io.output_writer as ow


class LIMIX_runner(object):
    '''
    A class for running a LIMIX pipeline.
    '''
    __slots__=["options","args","data", "result","infostring"]

    def __init__(self,infostring=None):
        '''
        nothing to initialize
        '''
        self.options=None
        self.args=None
        self.data=None
        self.result={}
        self.infostring=infostring
        if self.infostring is not None:
            self.result["infostring"]=self.infostring
        pass

    def parse_args(self):
        usage = "usage: %prog [options]"
        parser = OptionParser(usage=usage)
        parser.add_option("-D","--data_script", action="store", dest='data_script', type=str, help='A python script for loading the data', default=r"example_scripts/example_data.py")
        parser.add_option("-E","--experiment_script", action="store", dest='experiment_script', type=str, help='A python script for running the experiment', default=r"example_scripts/gwas_example.py")
        parser.add_option("-O","--outpath", action="store", dest='outpath', type=str, help='The output path (or file, if ending is .h5, the ooutput will be written to HDF5 file)', default="example_out")
        #parser.add_option('--seed', metavar='seed', type=int, help='The random seed', default=123123)
        parser.add_option("-T","--timestamp",action="store_true", dest='timestamp', help="Append unique timestamp to output value", default=False)
        parser.add_option("-R","--delimiter", action="store", dest='delimiter',type=str, help="The delimiter between output values", default="\t")
        parser.add_option("-F","--float_format", action="store", dest='float_format',type=str, help="Formating string for floating point output values", default="%.6e")
        (self.options, self.args) = parser.parse_args()
        self.result["options"]=str(self.options)
        return (self.options,self.args)

    def load_data(self):
        """
        Run the job specified in data_script
        """
        options=self.options
        command = open(self.options.data_script).read()
        self.result["data_script"]=command
        t0=time.time()
        data=None       #fallback data
        exec(command)   #creates variable data
        t1=time.time()
        print(("Elapsed time for data reading is %.2f seconds" % (t1-t0)))
        self.data=data
        return self.data

    def run_experiment(self):
        """
        Run the job specified in experiment_script
        """
        data=self.data
        options=self.options
        result=self.result

        command = open(self.options.experiment_script).read()
        result["experiment_script"]=command
        t0=time.time()
        exec(command)   #creates variable result
        t1=time.time()
        print(("Elapsed time for running the experiment is %.2f seconds" % (t1-t0)))
        self.result=result
        return self.result

    def write_resultfiles(self):
        """
        Write the output to disk
        """
        t0=time.time()
        writer = ow.output_writer(output_dictionary=self.result)
        if len(self.options.outpath)>=3 and self.options.outpath[-3:]==".h5":
            writer.write_hdf5(filename=self.options.outpath,timestamp=self.options.timestamp)
        else:
            writer.write_txt(outdir=self.options.outpath,timestamp=self.options.timestamp,delimiter=self.options.delimiter,float_format=self.options.float_format)
        t1=time.time()
        print(("Elapsed time for writing the output files is %.2f seconds" % (t1-t0)))


def entry_point():
    infostring = "limix_runner.py, Copyright(c) 2014, The LIMIX developers\nlast modified: %s" % time.ctime(os.path.getmtime(__file__))
    print (infostring)


    runner = LIMIX_runner(infostring=infostring)
    (options,args) = runner.parse_args()
    data = runner.load_data()
    result = runner.run_experiment()
    runner.write_resultfiles()



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# modified by Barbara Rakitsch
import pdb
import os
import numpy as SP
import csv

def standardize(snps): 
    mdat = SP.ma.masked_array(snps,SP.isnan(snps))  # mask nan values
    N = (mdat - SP.mean(mdat,axis=0)) / SP.std(mdat,axis=0)  # standardize column-wise to mean 0 and std 1
    return N.filled(0.0)  # assign 0.0 to masked missing values

def readBIM(basefilename,usecols=None):
    """
    helper method for speeding up read BED
    """
    bim = basefilename+ '.bim'
    bim = SP.loadtxt(bim,dtype=bytes,usecols=usecols)
    return bim
    

def readFAM(basefilename,usecols=None):
    """
    helper method for speeding up read FAM
    """
    fam = basefilename+'.fam'
    fam = SP.loadtxt(fam,dtype=bytes,usecols=usecols)
    return fam


def readBED(basefilename, useMAFencoding=False,blocksize = 1, start = 0, nSNPs = SP.inf, startpos = None, endpos = None, order  = 'F',standardizeSNPs=False,ipos = 2,bim=None,fam=None):
    '''
    read [basefilename].bed,[basefilename].bim,[basefilename].fam
    --------------------------------------------------------------------------
    Input:
    basefilename    : string of the basename of [basename].bed, [basename].bim,
                      and [basename].fam
    blocksize       : load blocksize SNPs at a time (default 1)
    start           : index of the first SNP to be loaded from the .bed-file
                      (default 0)
    nSNPs           : load nSNPs from the .bed file (default SP.inf, meaning all)
    startpos        : starting position of the loaded genomic region[chr,bpdist]
    endpos          : end-position of the loaded genomic region     [chr,bpdist]
    order           : memory layout of the returned SNP array (default 'F')
                      'F'   : Fortran-style column-major array (SNP-major)
                      'C'   : C-style row-major array (individual-major)
    standardizeSNPs : bool indeicator if the resulting SNP array is supposed to 
                      be zero-mean and unit-vatiance with mean imputed missing
                      values (default False)
    ipos            : the index of the position index to use (default 2)
                        1 : genomic distance
                        2 : base-pair distance
    useMAFencoding  : if set to one, the minor allele is encoded with 2, the major allele with 0.
                      otherwise, the plink coding is used (default False).
    --------------------------------------------------------------------------
    Output dictionary:
    'rs'     : [S] array rs-numbers
    'pos'    : [S*3] array of positions [chromosome, genetic dist, basepair dist]
    'snps'   : [N*S] array of snp-data
    'iid'    : [N*2] array of family IDs and individual IDs
    --------------------------------------------------------------------------    
    '''
    
    if bim is None: bim = readBIM(basefilename,usecols=(0,1,2,3))
    if fam is None: fam = readFAM(basefilename,usecols=(0,1))

    
    rs = bim[:,1]
    pos = SP.array(bim[:,(0,2,3)],dtype = 'float')


    
    if startpos is not None:
        #pdb.set_trace()
        i_c = pos[:,0]==startpos[0]
        i_largerbp = pos[:,ipos]>=startpos[ipos]
        start = which(i_c * i_largerbp)
        while (start-1 >= 0 and pos[start-1,ipos] == startpos[ipos]):
            start = start -1
        i_c = pos[:,0]==endpos[0]
        i_smallerbp = pos[:,ipos]>=endpos[ipos]
        end = which(i_c * i_smallerbp)
        while (end+1 < pos.shape[0] and pos[end+1,ipos] == endpos[ipos]):
            end = end + 1
        nSNPs = end - start
        if (nSNPs<=0) or (end==0) or (start<=0):
            ret = {
                'pos':SP.zeros((0,3)),
                'rs':SP.zeros((0)),
                'iid':fam,
                'snps':SP.zeros((fam.shape[0],0))
                }
            return ret
        pass
    N = fam.shape[0]
    S = bim.shape[0]
    S_res = min(S,start + nSNPs)
    nSNPs = min(S-start,nSNPs)
    #if startpos is not None:
	#print("start: " + str(start))
	#print("end: " + str(end))
    #print("S_res: " + str(S_res))
    #print("nSNPs: " + str(nSNPs))
    if nSNPs<=0:
        ret = {
            'rs'     :rs[start:start],
            'pos'    :pos[start:start,:],
            #'snps'   :SNPs[0:N,start:start],
            'snps'   :SP.zeros((N,0)),
            'iid'    :fam
            }
        return ret
    SNPs = SP.zeros(((SP.ceil(0.25*N)*4),nSNPs),order=order)
    bed = basefilename + '.bed'
    with open(bed, "rb") as f:
        mode = f.read(2)
        if mode != b'l\x1b':
            raise Exception('No valid binary PED file')
        mode = f.read(1) #\x01 = SNP major \x00 = individual major
        if mode != b'\x01':
            raise Exception('only SNP-major is implemented')
        startbit = SP.ceil(0.25*N)*start+3
        f.seek(int(startbit))
        for blockStart in SP.arange(0,nSNPs,blocksize, dtype=int):
            blockEnd = int(min(S,blockStart+blocksize))
            Sblock = min(nSNPs-blockStart,blocksize)
            nbyte = int(SP.ceil(0.25*N)*Sblock)
            bytes = SP.array(bytearray(f.read(nbyte))).reshape((SP.ceil(0.25*N),Sblock),order='F')
            
            SNPs[3::4,blockStart:blockEnd][bytes>=64]=SP.nan
            SNPs[3::4,blockStart:blockEnd][bytes>=128]=1
            SNPs[3::4,blockStart:blockEnd][bytes>=192]=2
            bytes=SP.mod(bytes,64)
            SNPs[2::4,blockStart:blockEnd][bytes>=16]=SP.nan
            SNPs[2::4,blockStart:blockEnd][bytes>=32]=1
            SNPs[2::4,blockStart:blockEnd][bytes>=48]=2
            bytes=SP.mod(bytes,16)
            SNPs[1::4,blockStart:blockEnd][bytes>=4]=SP.nan
            SNPs[1::4,blockStart:blockEnd][bytes>=8]=1
            SNPs[1::4,blockStart:blockEnd][bytes>=12]=2
            bytes=SP.mod(bytes,4)
            SNPs[0::4,blockStart:blockEnd][bytes>=1]=SP.nan
            SNPs[0::4,blockStart:blockEnd][bytes>=2]=1
            SNPs[0::4,blockStart:blockEnd][bytes>=3]=2
    
    if 0: #the binary format as described in the documentation (seems wrong)
        SNPs[3::4][bytes>=128]=SP.nan
        SNPs[3::4][bytes>=192]=1
        bytes=SP.mod(bytes,128)
        SNPs[3::4][bytes>=64]+=1
        bytes=SP.mod(bytes,64)
        SNPs[2::4][bytes>=32]=SP.nan
        SNPs[2::4][bytes>=48]=1
        bytes=SP.mod(bytes,32)
        SNPs[2::4][bytes>=16]+=1
        bytes=SP.mod(bytes,16)
        SNPs[1::4][bytes>=8]=SP.nan
        SNPs[1::4][bytes>=12]=1
        bytes=SP.mod(bytes,8)
        SNPs[1::4][bytes>=4]+=1
        bytes=SP.mod(bytes,4)
        SNPs[0::4][bytes>=2]=SP.nan
        SNPs[0::4][bytes>=3]=1
        bytes=SP.mod(bytes,2)
        SNPs[0::4][bytes>=1]+=1
    snps = SNPs[0:N,:]

    if useMAFencoding:
        imaf = SP.sum(snps==2,axis=0)>SP.sum(snps==0,axis=0)
        snps[:,imaf] = 2 - snps[:,imaf]
        
    if standardizeSNPs:
        snps = standardize(snps)
    ret = {
            'rs'     :rs[start:S_res],
            'pos'    :pos[start:S_res,:],
            'snps'   :snps,
            'iid'    :fam
            }
    return ret




import numpy as np
import scipy as sp
import scipy.stats

def _calculate_number_alleles(G):
    G = np.asarray(G, int)
    assert len(G.shape) == 2
    u = np.unique(G[:])
    assert np.all([ui in [0, 1, 2] for ui in u])

    b = np.sum(G, axis=0)
    a = G.shape[0]*2 - b

    return (a, b)

def _normalize_maf_allele(G):
    (a, b) = _calculate_number_alleles(G)
    change = b > a
    G[:, change] = 2 - G[:, change]

def _calculate_maf(G):
    return np.sum(G, 0) / float(2*G.shape[0])

# According to GCTA's paper
def grm_unbiased(G):
    _normalize_maf_allele(G)
    p = _calculate_maf(G)
    denom = 2 * p * (1. - p)
    n = G.shape[0]

    K = np.zeros((n, n))

    for j in range(n-1):
        for k in range(j+1, n):

            v0 = (G[j, :] - 2*p) * (G[k, :] - 2*p)

            K[j, k] = np.mean(v0 / denom)
            K[k, j] = K[j, k]

    for j in range(n):
        g2 = G[j, :]**2
        v0 = g2 - (1 + 2*p) * G[j, :] + 2 * p**2
        K[j, j] = 1 + np.mean(v0 / denom)

    return K

if __name__ == '__main__':
    # np.random.seed(5)
    # # G = np.random.randint(0, 3, (100000, 1))
    # G = sp.stats.binom.rvs(2, 0.5, size=(1000, 10))
    # # _calculate_maf(G)
    # K = grm_unbiased(G)
    # ; ipdb.set_trace()
    # print K


    import numpy as np
    import scipy as sp
    import scipy.stats
    np.random.seed(0)
    N = 5

    nfrX = sp.stats.binom.rvs(2, 0.3, size=(N, 10))
    nbgX = sp.stats.binom.rvs(2, 0.5, size=(N, 10))
    y = np.random.randint(0, 2, size=N)
    # r = apply_gcta(nfrX, nbgX, y, 0.5)

    K = grm_unbiased(nbgX)
    print((np.diagonal(K)))
    # diag 0.8817461 0.9085317 0.6531746 1.2656746 0.5007936
    # print np.mean([0.8817461, 0.9085317, 0.6531746, 1.2656746, 0.5007936])

    #  [1] -0.06626985  0.15158729 -0.44087306 -0.62341273 -0.30337301 -0.32182536
    #  [7] -0.34365076 -0.09801586 -0.04206349 -0.01706349



import unittest
from limix.core.covar import FreeFormCov
from limix.core.mean import MeanKronSum
from limix.core.gp import GP3KronSumLR
from limix.core.gp import GP
from limix.utils.preprocess import covar_rescale
from limix.utils.check_grad import mcheck_grad

import numpy as np
import scipy as sp
import pdb


class TestGPBase(unittest.TestCase):
    def setUp(self):
        np.random.seed(1)

        # define phenotype
        N = 200
        P = 2
        Y = sp.randn(N,P)
        # define row caoriance
        f = 10
        G = 1.*(sp.rand(N, f)<0.2)
        X = 1.*(sp.rand(N, f)<0.2)
        R = covar_rescale(sp.dot(X,X.T))
        R+= 1e-4 * sp.eye(N)
        # define col covariances
        Cg = FreeFormCov(P)
        self._Cg = Cg
        Cn = FreeFormCov(P)
        Cg.setCovariance(0.5 * sp.cov(Y.T))
        Cn.setCovariance(0.5 * sp.cov(Y.T))
        # define gp
        self.gp = GP3KronSumLR(Y = Y, Cg = Cg, Cn = Cn, R = R, G = G, rank = 1)

    def test_grad(self):

        gp = self.gp

        def func(x, i):
            params = gp.getParams()
            params['covar'] = x
            gp.setParams(params)
            return gp.LML()

        def grad(x, i):
            params = gp.getParams()
            params['covar'] = x
            gp.setParams(params)
            grad = gp.LML_grad()
            return grad['covar'][i]

        x0 = gp.getParams()['covar']
        err = mcheck_grad(func, grad, x0)
        np.testing.assert_almost_equal(err, 0., decimal=4)

    def test_grad_activation(self):

        gp = self.gp

        self._Cg._K_act = False

        def func(x, i):
            params = gp.getParams()
            params['covar'] = x
            gp.setParams(params)
            return gp.LML()

        def grad(x, i):
            params = gp.getParams()
            params['covar'] = x
            gp.setParams(params)
            grad = gp.LML_grad()
            return grad['covar'][i]

        x0 = gp.getParams()['covar']
        err = mcheck_grad(func, grad, x0)
        np.testing.assert_almost_equal(err, 0., decimal=4)

if __name__ == "__main__":
    unittest.main()



import scipy as SP
import os

def readBimFile(basefilename):
    """
    Helper fuinction that reads bim files
    """
    # read bim file
    bim_fn = basefilename+'.bim'
    rv = SP.loadtxt(bim_fn,delimiter='\t',usecols = (0,3),dtype=int)
    return rv

def readCovarianceMatrixFile(cfile,readCov=True,readEig=True):
    """"
    reading in similarity matrix

    cfile   File containing the covariance matrix. The corresponding ID file must be specified in cfile.id)
    """
    covFile = cfile+'.cov'
    evalFile = cfile+'.cov.eval'
    evecFile = cfile+'.cov.evec'

    RV = {}
    if readCov:
        assert os.path.exists(covFile), '%s is missing.'%covFile
        RV['K'] = SP.loadtxt(covFile)
    if readEig:
        assert os.path.exists(evalFile), '%s is missing.'%evalFile
        assert os.path.exists(evecFile), '%s is missing.'%evecFile
        RV['eval'] = SP.loadtxt(evalFile)
        RV['evec'] = SP.loadtxt(evecFile)

    return RV


def readCovariatesFile(fFile):
    """"
    reading in covariate file

    cfile   file containing the fixed effects as NxP matrix
            (N=number of samples, P=number of covariates)
    """
    assert os.path.exists(fFile), '%s is missing.'%fFile
    F = SP.loadtxt(fFile)
    if F.ndim==1: F=F[:,SP.newaxis]
    return F


def readPhenoFile(pfile,idx=None):
    """"
    reading in phenotype file

    pfile   root of the file containing the phenotypes as NxP matrix
            (N=number of samples, P=number of traits)
    """

    usecols = None
    if idx!=None:
        """ different traits are comma-seperated """
        usecols = [int(x) for x in idx.split(',')]
        
    phenoFile = pfile+'.phe'
    assert os.path.exists(phenoFile), '%s is missing.'%phenoFile
    Y = SP.loadtxt(phenoFile,usecols=usecols)
    if (usecols is not None) and (len(usecols)==1): Y = Y[:,SP.newaxis]
        
    Y -= Y.mean(0); Y /= Y.std(0)
    return Y

def readNullModelFile(nfile):
    """"
    reading file with null model info

    nfile   File containing null model info
    """

    params0_file = nfile+'.p0'
    nll0_file = nfile+'.nll0'
    assert os.path.exists(params0_file), '%s is missing.'%params0_file
    assert os.path.exists(nll0_file), '%s is missing.'%nll0_file
    params = SP.loadtxt(params0_file)
    NLL0 = SP.array([float(SP.loadtxt(nll0_file))])


    if params.ndim==1:
        rv = {'params0_g':SP.array([params[0]]),'params0_n':SP.array([params[1]]),'NLL0':NLL0}
    else:
        rv = {'params0_g':params[0],'params0_n':params[1],'NLL0':NLL0}

    return rv

def readWindowsFile(wfile):
    """"
    reading file with windows

    wfile   File containing window info
    """
    window_file = wfile+'.wnd'
    assert os.path.exists(window_file), '%s is missing.'%window_file
    rv = SP.loadtxt(window_file)
    return rv




'''
Created on Dec 19, 2013

@author: johannes stephan, barbara rakitsch
'''

from sklearn.linear_model import Lasso
from sklearn.metrics import mean_squared_error
import sklearn.cross_validation as cross_validation

from .varianceDecomposition import VarianceDecomposition
from . import qtl
import scipy as SP
import numpy as NP
import scipy.stats as ST
import time
import scipy.linalg as LA
import scipy.optimize as OPT
import pdb


def compute_linear_kernel(X,idx=None,jitter=1e-3,standardize=True):
    """
    compute linear kernel

    X           : SNP data [N x F]
    idx         : boolean vector of size F, indicating which SNPs to use to build the covariance matrix
    standardize : if True (default), covariance matrix is standardized to unit variance
    jitter      : adds jitter to the diagonal of the covariance matrix (default: 1e-3)
    """
    N = X.shape[0]
    if idx is not None:
        K = SP.dot(X[:,idx],X[:,idx].T)
    else:
        K = SP.dot(X,X.T)

    if standardize:
        K /= SP.diag(K).mean()

    if jitter:
        K += jitter*SP.eye(N)

    return K


def runStabilitySelection(estimator,X,y,K=None,n_repeats=100,frac_sub=0.5,verbose=False):
    """
    estimator   : (Lmm-)Lasso model
    X           : SNP data
    y           : phenotype vector
    K           : background covariance matrix
    n_repeats   : number of repetitions
    frac_sub    : size of the subsample
    """
    N,F = X.shape
    Nsub = int(frac_sub*N)

    W_nonzero = SP.zeros((n_repeats,F),dtype=bool)

    for irep in range(n_repeats):
        if verbose: print(('Running %d/%d'%(irep,n_repeats)))
        iperm = SP.random.permutation(N)[:Nsub]

        if K is not None:
            estimator.fit(X[iperm],y[iperm],K=K[iperm][:,iperm])
        else:
            estimator.fit(X[iperm],y[iperm])

        W_nonzero[irep] = estimator.coef_!=0

    p_stab = W_nonzero.mean(0)
    return p_stab


def testLassoBackgroundModel(estimator,Xfg,Xbg,y,use_Kgeno=True,**kwargs):
    """
    run a linear mixed model on the foreground SNPs (Xfg), while estimating SNPs for the background covariance matrix via the Lasso usting the background SNPs(Xbg).

    estimator   : (Lmm-)Lasso model
    Xbg         : SNP data for the background model
    Xfg         : SNP data for the foreground model
    y           : phenotype vector
    use_Kgene   : if True (default), background covariance matrix is estimated and used.
    """
    if use_Kgeno:
        Kbg = compute_linear_kernel(Xbg)
        estimator.fit(Xfg,y,Kbg)
        iactive = estimator.coef_!=0

        if iactive.any():
            Kfg = compute_linear_kernel(Xbg,iactive)
            vd = VarianceDecomposition(y)
            vd.addRandomEffect(is_noise=True)
            vd.addRandomEffect(Kbg)
            vd.addRandomEffect(Kfg)
            vd.optimize()
            K = vd.gp.getCovar().K()
        else:
            K = Kbg

    else:
        estimator.fit(Xbg,y)
        idx = estimator.coef_!=0
        K   = compute_linear_kernel(Xbg,idx)

    lmm = qtl.test_lmm(Xfg,y,K=K,**kwargs)
    pv  = lmm.getPv()

    return pv

def runCrossValidation(estimator,X,y,alphas,K=None,n_folds=10,verbose=False):
    """ run cross-validation

    estimator   : (Lmm-)Lasso model
    X           : SNP data
    y           : phenotype vector
    alphas      : list of l1-regularization parameter
    K           : background covariance matrix
    n_folds     : number of folds used in cross-validation
    """

    """ setting up """
    N = X.shape[0]
    kfold = cross_validation.KFold(N, n_folds=n_folds)
    n_alphas = len(alphas)
    MSE_train = SP.zeros((n_folds,n_alphas))
    MSE_test  = SP.zeros((n_folds,n_alphas))
    W_nonzero = SP.zeros((n_folds,n_alphas))

    t1 = time.time()
    ifold = 0
    for train,test in kfold:
        if verbose: print(('running fold %d'%ifold))

        """ splitting into training and testing """
        X_train = X[train]
        X_test  = X[test]
        y_train = y[train]
        y_test  = y[test]
        if K is not None:
            K_train = K[train][:,train]
            K_test  = K[test][:,train]

        for ialpha in range(n_alphas):
            estimator.set_params(alpha=alphas[ialpha])
            if K is None:
                estimator.fit(X_train,y_train)
                ytrain_star = estimator.predict(X_train)
                ytest_star  = estimator.predict(X_test)
            else:
                estimator.fit(X_train,y_train,K_train)
                ytrain_star = estimator.predict(X_train,K_train)
                ytest_star  = estimator.predict(X_test, K_test)

            MSE_train[ifold,ialpha]  = mean_squared_error(ytrain_star,y_train)
            MSE_test[ifold,ialpha]   = mean_squared_error(ytest_star,y_test)
            W_nonzero[ifold,ialpha]  = SP.sum(estimator.coef_!=0)

        ifold +=1

    t2 = time.time()
    if verbose: print(('finished in %.2f seconds'%(t2-t1)))
    return MSE_train,MSE_test,W_nonzero




class LmmLasso(Lasso):
    """
    Lmm-Lasso classo
    """
    def __init__(self, alpha=1., **lasso_args):
        """
        Extension to the sklearn's LASSO to model population structure

        alpha: l1-regularization parameter
        """
        super(LmmLasso, self).__init__(alpha=alpha, **lasso_args)
        self.msg = 'lmmlasso'




    def fit(self, X, y, K, standardize=False, verbose=False,**lasso_args):
        """
        fitting the model

        X: SNP data
        y: phenotype data
        K: backgroundcovariance matrix
        standardize: if True, genotypes and phenotypes are standardized
        """

        if y.ndim == 2:
            assert y.shape[1]==1, 'Only one phenotype can be processed at at time.'
            y = y.flatten()
        time_start = time.time()
        [n_s, n_f] = X.shape
        assert X.shape[0] == y.shape[0], 'dimensions do not match'
        assert K.shape[0] == K.shape[1], 'dimensions do not match'
        assert K.shape[0] == X.shape[0], 'dimensions do not match'

        """ standardizing genotypes and phenotypes """
        if standardize:
            X -= X.mean(axis=0)
            X /= X.std(axis=0)
            y -= y.mean(axis=0)
            y /= y.std(axis=0)

        """ training null model """
        vd = VarianceDecomposition(y)
        vd.addRandomEffect(is_noise=True)
        vd.addRandomEffect(K)
        vd.optimize()
        varComps = vd.getVarianceComps()
        delta0   = varComps[0,0]/varComps.sum()
        self.varComps = varComps

        S,U = LA.eigh(K)

        """ rotating data """
        Sdi = 1. / (S + delta0)
        Sdi_sqrt = SP.sqrt(Sdi)
        SUX = SP.dot(U.T, X)
        SUX = SUX * SP.tile(Sdi_sqrt, (n_f, 1)).T
        SUy = SP.dot(U.T, y)
        SUy = Sdi_sqrt * SUy

        """ fitting lasso """
        super(LmmLasso, self).fit(SUX, SUy, **lasso_args)
        yhat = super(LmmLasso, self).predict(X)
        self.w_ridge = LA.solve(K + delta0 * SP.eye(n_s), y - yhat)

        time_end = time.time()
        time_diff = time_end - time_start
        if verbose: print(('... finished in %.2fs'%(time_diff)))

        return self

    def getVarianceComps(self,univariance=False):
        """
        Return the estimated variance components

        Args:
            univariance:   Boolean indicator, if True variance components are normalized to sum up to 1 for each trait
        Returns:
            variance components of all random effects [noise, signal]
        """
        if univariance:
            self.varComps /= self.varComps.mean()

        return self.varComps

    def predict(self, Xstar, Kstar=None):
        """
        predicting the phenotype

        Xstar: SNP data
        Kstar: covariance matrix between test and training samples
        """

        assert self.w_ridge.shape[0]==Kstar.shape[1], 'number of training samples is not consistent.'
        assert self.coef_.shape[0]==Xstar.shape[1],   'number of SNPs is not consistent.'
        assert Xstar.shape[0]==Kstar.shape[0],   'number of test samples is not consistent.'

        fixed_effect = super(LmmLasso, self).predict(Xstar)
        if Kstar is not None:
            return fixed_effect + SP.dot(Kstar, self.w_ridge)
        else:
            return fixed_effect



import scipy as sp
import h5py
#re for regular expressions
import re
import pdb

class SocialData():
    
    def __init__(self, task = None, kinship_type = "", subset = None):
        assert task is not None, 'Specify task!'
        self.task=task
        self.kinship_type=kinship_type
        self.subset = subset
        self.load()
    
    
    def load(self):
        
        if self.task == 'HSmice_data_REVISIONS':
            in_file = '/nfs/leia/research/stegle/abaud/HSmice/data/HSmice_wPleth_REVISIONS.hdf5'
            f = h5py.File(in_file,'r')
            
            self.measures = f['data_bcNcovariates']['cols_measures']['measures'][:]
            self.all_pheno = f['data_bcNcovariates']['array'][:].T
            self.pheno_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]
            self.all_pheno = f['data_bcNcovariates']['array'][:].T
            self.pheno_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]
            self.all_covs2use = f['data_bcNcovariates']['cols_measures']['covariatesUsed']
            self.all_covs = f['data_bcNcovariates']['array'][:].T
            self.covs_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]

            if self.kinship_type == 'old_IBS':
                self.kinship_full = f['kinships']['old_IBS']['array'][:]
                self.kinship_full_ID = f['kinships']['old_IBS']['cols_subjects'][:]
            elif self.kinship_type == 'Andres_kinship':
                self.kinship_full = f['kinships']['Andres_kinship']['array'][:]
                self.kinship_full_ID = f['kinships']['Andres_kinship']['cols_subjects'][:]
            self.cage_full = f['data_bcNcovariates']['rows_subjects']['cage'][:]
            self.cage_full_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]
            if self.subset is not None:
                self.subset_IDs = f['subsets'][self.subset][:]
            else:
                self.subset_IDs = None

        elif self.task == 'HSmice_hip_REVISIONS':
            in_file = '/nfs/leia/research/stegle/abaud/HSmice/data/HSmice_wPleth_REVISIONS.hdf5'
            f = h5py.File(in_file,'r')
            
            self.measures = f['hip_expr']['cols_probes']['probes'][:]
            self.all_pheno = f['hip_expr']['array'][:].T
            self.pheno_ID = f['hip_expr']['rows_subjects'][:]
            self.all_covs = f['data_bcNcovariates']['array'][:].T
            self.covs_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]
            if self.kinship_type == 'old_IBS':
                self.kinship_full = f['kinships']['old_IBS']['array'][:]
                self.kinship_full_ID = f['kinships']['old_IBS']['cols_subjects'][:]
            elif self.kinship_type == 'Andres_kinship':
                self.kinship_full = f['kinships']['Andres_kinship']['array'][:]
                self.kinship_full_ID = f['kinships']['Andres_kinship']['cols_subjects'][:]
            self.cage_full = f['data_bcNcovariates']['rows_subjects']['cage'][:]
            self.cage_full_ID = f['data_bcNcovariates']['rows_subjects']['outbred'][:]
            if self.subset is not None:
                self.subset_IDs = f['subsets'][self.subset][:]
            else:
                self.subset_IDs = None


        elif 'HSmice_simulations' in self.task:
            if self.task == 'HSmice_simulations_realized':
                in_file = '/nfs/research/stegle/projects/IGE_outbreds/HSmice/simulations/revisions_HSmice_simulations_realized.hdf5'
            elif self.task == 'HSmice_simulations_all':
                in_file = '/nfs/research/stegle/projects/IGE_outbreds/HSmice/simulations/negcorr_revisions_HSmice_simulations_around_best.hdf5'
            f = h5py.File(in_file,'r')
    
            self.measures = f['simulations']['cols_measures']['measures'][:]
            self.all_pheno = f['simulations']['array'][:].T
            self.pheno_ID = f['simulations']['rows_subjects']['outbred'][:]
            self.covs_ID = None

            in_file = '/nfs/research/stegle/projects/IGE_outbreds/HSmice/HSmice_wPleth_new.hdf5'
            f = h5py.File(in_file,'r')

            if self.kinship_type == 'Andres_kinship':
                self.kinship_full = f['kinships']['Andres_kinship']['array'][:]
                self.kinship_full_ID = f['kinships']['Andres_kinship']['cols_subjects'][:]
            self.cage_full = f['residualsNcovariates']['rows_subjects']['cage'][:]
            self.cage_full_ID = f['residualsNcovariates']['rows_subjects']['outbred'][:]
            if self.subset is not None:
                self.subset_IDs = f['subsets'][self.subset][:]
            else:
                self.subset_IDs = None

        else:
            print("Nothing done: task unknown!")


    def get_data(self,col):
        
        if self.task == 'HSmice_data_REVISIONS':
            self.trait=self.measures[col]
            self.pheno = self.all_pheno[:,col]
            covs2use = self.all_covs2use[col].split(',')
            Ic = sp.zeros(self.measures.shape[0],dtype=bool)
            for cov in covs2use:
                Ic = sp.logical_or(Ic,self.measures==cov)
                self.covs = self.all_covs[:,Ic]

        elif self.task == 'HSmice_hip_REVISIONS':
            self.trait=self.measures[col]
            self.pheno = self.all_pheno[:,col]
            covs2use = ['GENDER','group_size']
            Ic = sp.zeros(self.measures.shape[0],dtype=bool)
            for cov in covs2use:
                Ic = sp.logical_or(Ic,self.measures==cov)
                self.covs = self.all_covs[:,Ic]

        elif 'HSmice_simulations' in self.task:
            self.trait = self.measures[col]
            self.pheno = self.all_pheno[:,col]
            self.covs = None

        else:
            print("Nothing done: task unknown!")


        return {'trait' : self.trait,
                'pheno' : self.pheno,
                'pheno_ID' : self.pheno_ID,
                'covs' : self.covs,
                'covs_ID' : self.covs_ID,
                'kinship_type' : self.kinship_type,
                'kinship_full' : self.kinship_full,
                'kinship_full_ID' : self.kinship_full_ID,
                'cage_full' : self.cage_full,
                'cage_full_ID' : self.cage_full_ID,
                'subset_IDs' : self.subset_IDs}









import scipy as sp

def vei_CoR_veX(X, C=None, R=None):
    """
    Args:
        X:  NxPxS tensor
        C:  CxC row covariance (if None: C set to I_PP)
        R:  NxN row covariance (if None: R set to I_NN)
    Returns:
        NxPxS tensor obtained as ve^{-1}((C \kron R) ve(X))
        where ve(X) reshapes X as a NPxS matrix.
    """
    _X = X.transpose((0,2,1))
    if R is not None:   RV = sp.tensordot(R, _X, (1,0))
    else:               RV = _X
    if C is not None:   RV = sp.dot(RV, C.T)
    return RV.transpose((0,2,1))

